# cython: language_level=3

# TRX Toolkit
# Simple TDMA frame clock generator
#
# (C) 2017-2019 by Vadim Yanitskiy <axilirator@gmail.com>
#
# All Rights Reserved
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

import logging as log
import time
import os
import sys

from gsm_shared import *


ns = 1e-9
us = 1e-6


class CLCKGen:
	# GSM TDMA definitions
	SEC_DELAY_US = 1000 * 1000
	GSM_FRAME_US = 4615.0

	def __init__(self, clck_links, clck_start = 0, ind_period = 102):
		self.clck_links = clck_links
		self.ind_period = ind_period
		self.clck_start = clck_start

		# Calculate counter time
		self.ctr_interval  = self.GSM_FRAME_US
		self.ctr_interval /= self.SEC_DELAY_US
		self._t_tick  = int(self.ctr_interval // ns)

		# Initialize timer fd
		self._timerfd = _os_timerfd_create(time.CLOCK_MONOTONIC)

		# (Optional) clock consumer
		self.clck_handler = None

	def __del__(self):
		os.close(self._timerfd)

	@property
	def running(self):
		t_next, _ = _os_timerfd_gettime_ns(self._timerfd)
		return (t_next != 0)

	def start(self):
		# (Re)set the clock counter
		self.clck_src = self.clck_start

		# start timer fd
		_os_timerfd_settime_ns(self._timerfd, flags=0, initial=self._t_tick, interval=self._t_tick)

	def stop(self):
		# stop timer fd
		_os_timerfd_settime_ns(self._timerfd, flags=0, initial=0, interval=0)

	# tick must be called periodically by CLCKGen user.
	#
	# It waits for the next GSM frame to happen and emits corresponding clock indication at that time.
	# It also runs attached .clck_handler if there is one.
	#
	# It is possible to use .tick in both blocking and non-blocking ways:
	#
	# - without extra care .tick will block waiting for the next GSM frame as explained above,
	# - client code can also poll/select on ._timerfd to wait for GSM frame.
	#   After ._timerfd becomes ready it is guaranteed that the next .tick call will not block.
	def tick(self):
		# run .send_clck_ind() every .ctr_interval
		# NOTE timerfd is careful not to accumulate timing error when organizing the clock loop

		_ = os.read(self._timerfd, 8)
		assert len(_) == 8, len(_)
		ticks = int.from_bytes(_, byteorder=sys.byteorder)
		assert ticks > 0, ticks
		if ticks > 1:
			log.warning("CLCKGen: time overrun by %dus; resetting the clock" % ((ticks-1)*self._t_tick * ns // us))
			# (the kernel does clock reset by itself)

		self.send_clck_ind()

	def send_clck_ind(self):
		# We don't need to send so often
		if self.clck_src % self.ind_period == 0:
			# Create UDP payload
			payload = "IND CLOCK %u\0" % self.clck_src

			# Send indication to all UDP links
			for link in self.clck_links:
				link.send(payload)

			# Debug print
			log.debug(payload.rstrip("\0"))

		if self.clck_handler is not None:
			self.clck_handler(self.clck_src)

		# Increase frame count (modular arithmetic)
		self.clck_src = (self.clck_src + 1) % GSM_HYPERFRAME


# ---- misc ----

# _os_timerfd_<func> is like os.timerfd_<func> but works on all python versions.
# we do it ourselves because os.timerfd_* is available only for py ≥ 3.13

from posix.time cimport timespec, itimerspec

cdef extern from "<sys/timerfd.h>":
	int timerfd_create(int clockid, int flags)
	int timerfd_settime(int fd, int flags, const itimerspec *new_value, itimerspec *old_value)
	int timerfd_gettime(int fd, itimerspec *curr_value)


cdef int _os_timerfd_create(int clockid, int flags=0) except -1:
	fd = timerfd_create(clockid, flags)
	if fd == -1:
		_raise_oserr()
	return fd

cdef _os_timerfd_settime_ns(int fd, flags=0, initial=0, interval=0):
	cdef itimerspec its
	its.it_value    .tv_sec  = initial  // 1000000000  # 1e9
	its.it_value    .tv_nsec = initial   % 1000000000
	its.it_interval .tv_sec  = interval // 1000000000
	its.it_interval .tv_nsec = interval  % 1000000000
	err = timerfd_settime(fd, flags, &its, NULL)
	if err == -1:
		_raise_oserr()

cdef _os_timerfd_gettime_ns(int fd):  # -> (next_expiration, interval)
	cdef itimerspec its
	err = timerfd_gettime(fd, &its)
	if err == -1:
		_raise_oserr()
	cdef object t_next     = 0L  # py long to avoid int overflow
	cdef object t_interval = 0L
	t_next += its.it_value.tv_sec
	t_next *= 1000000000  # 1e9
	t_next += its.it_value.tv_nsec
	t_interval += its.it_interval.tv_sec
	t_interval *= 1000000000  # 1e9
	t_interval += its.it_interval.tv_nsec
	return (t_next, t_interval)
